import asyncio
import re

from bs4 import BeautifulSoup
from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, write_last_post_file
from ..base_functions.data_classes import AuthData, CascadeItem
from ..client.client import Session


class XBunkerCrawler:
    def __init__(self, *, include_id=False, auth: AuthData = None, scraping_mapper, separate_posts=False, output_last=[False, None]):
        self.include_id = include_id
        self.separate_posts = separate_posts
        self.output_last = output_last
        self.username, self.password = (auth.username, auth.password) if auth else (None, None)
        self.scraping_mapper = scraping_mapper
        self.lock = 0

    async def login(self, session: Session, url: URL):
        domain = URL("https://" + url.host) / "login"
        text = await session.get_text(domain)
        if "You are already logged in" in text:
            return
        soup = BeautifulSoup(text, 'html.parser')

        inputs = soup.select('form input')
        data = {
            elem['name']: elem['value']
            for elem in inputs
            if elem.get('name') and elem.get('value')
        }
        data.update({"login": self.username, "password": self.password, "_xfRedirect": str(domain)})
        return await session.post_data_no_resp(domain/"login", data=data)

    async def fetch(self, session: Session, url: URL):
        await log("Starting scrape of " + str(url))
        cascade = CascadeItem({})

        try:
            if self.username and self.password:
                while True:
                    if self.lock == 0:
                        self.lock = 1
                        await self.login(session, url)
                        self.lock = 0
                        break
                    await asyncio.sleep(2)
            else:
                await log("login wasn't provided, consider using --leakednudes-username and --leakednudes-password")
                await log("Not being logged in might cause issues.")
            await self.parse_thread(session, url, cascade, "")
        except Exception as e:
            self.lock = 0
            await log(f"there was an error signing into {url.host}")
            await log(e)
            return

        await log("Finished scrape of " + str(url))
        return cascade

    async def parse_thread(self, session: Session, url: URL, cascade: CascadeItem, title: str):
        try:
            soup = await session.get_BS4(url)

            domain = URL("https://" + url.host)

            title_block = soup.select_one("h1[class=p-title-value]")
            for elem in title_block.find_all("a"):
                elem.decompose()

            if title:
                pass
            else:
                title = title_block.text
                if self.include_id:
                    titlep2 = url.name
                    titlep2 = [s for s in titlep2 if "." in s][-1]
                    title = title + " - " + titlep2
                title = await make_title_safe(title.replace(r"\n", "").strip())

            content_links = []

            post_number = str(url).split("post-")
            post_number = int(post_number[-1].strip("/")) if len(post_number) == 2 else None

            posts = soup.select("div[class='message-main uix_messageContent js-quickEditTarget']")
            for post in posts:
                post_num_str = post.select_one("li[class=u-concealed] a").get('href').split('/')[-1]
                post_num_int = int(post_num_str.split('post-')[-1])
                if post_number:
                    if post_number > post_num_int:
                        continue

                temp_title = title+"/"+post_num_str if self.separate_posts else title

                for elem in post.find_all('blockquote'):
                    elem.decompose()
                post_content = post.select_one("div[class=bbWrapper]")

                # Content links
                links = post_content.select('a')
                for link in links:
                    link = link.get('href')
                    if link.endswith("/"):
                        link = link[:-1]
                    if link.startswith('//'):
                        link = "https:" + link
                    elif link.startswith('/'):
                        link = domain / link[1:]
                    content_links.append([URL(link), temp_title])

                links = post.select("div[class='bbImageWrapper js-lbImage']")
                for link in links:
                    link = link.get('data-src')
                    if link.endswith("/"):
                        link = link[:-1]
                    if link.startswith('/'):
                        link = domain / link[1:]
                    content_links.append([URL(link), temp_title])

                links = post.select("div[class='bbImageWrapper lazyload js-lbImage']")
                for link in links:
                    link = link.get('data-src')
                    if link.endswith("/"):
                        link = link[:-1]
                    if link.startswith('//'):
                        link = "https:" + link
                    elif link.startswith('/'):
                        link = domain / link[1:]
                    content_links.append([URL(link), temp_title])

                links = post.select("video source")
                for link in links:
                    link = link.get('src')
                    if link.endswith("/"):
                        link = link[:-1]
                    if link.startswith('//'):
                        link = "https:" + link
                    elif link.startswith('/'):
                        link = domain / link[1:]
                    content_links.append([URL(link), temp_title])

                links = post.select('iframe[class=saint-iframe]')
                for link in links:
                    link = link.get('src')
                    content_links.append([URL(link), temp_title])

                attachments_block = post.select_one("section[class=message-attachments]")
                links = attachments_block.select("a[class='file-preview js-lbImage']") if attachments_block else []
                for link in links:
                    link = link.get('href')
                    if link.endswith("/"):
                        link = link[:-1]
                    if link.startswith('/'):
                        link = domain / link[1:]
                    in_prog_title = temp_title + "/Attachments"
                    await cascade.add_to_album(url.host, in_prog_title, URL(link), url)

                links = post_content.select("span[data-s9e-mediaembed-iframe]")
                for link in links:
                    embed_data = link.get("data-s9e-mediaembed-iframe")
                    embed_data = embed_data.replace("\/\/", "https://www.")
                    embed_data = embed_data.replace("\\", "")

                    embed_url = re.search(r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,4}\\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)", embed_data)
                    if embed_url:
                        embed_url = URL(embed_url.group(
                            0).replace("www.", ""))
                        content_links.append([embed_url, temp_title])

            forum_direct_urls = [x for x in content_links if x[0].host in url.host]
            content_links = [x for x in content_links if x not in forum_direct_urls]
            for link_title_bundle in forum_direct_urls:
                link = link_title_bundle[0]
                temp_title = link_title_bundle[1]
                in_prog_title = temp_title + "/Attachments"
                if str(link).endswith("/"):
                    link = URL(str(link)[:-1])
                if 'attachments' in link.parts:
                    await cascade.add_to_album(url.host, in_prog_title, link, url)
                if 'content' in link.parts:
                    await cascade.add_to_album(url.host, in_prog_title, link, url)
                elif 'data' in link.parts:
                    await cascade.add_to_album(url.host, in_prog_title, link, url)

            tasks = []
            for link_title_bundle in content_links:
                link = link_title_bundle[0]
                temp_title = link_title_bundle[1]
                tasks.append(self.scraping_mapper.map_url(link, temp_title))
            await asyncio.gather(*tasks)

            next_page = soup.select_one('a[class="pageNav-jump pageNav-jump--next"]')
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    if next_page.startswith('/'):
                        next_page = domain / next_page[1:]
                    next_page = URL(next_page)
                    title = await self.parse_thread(session, next_page, cascade, title)

            if self.output_last[0]:
                if 'page-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                elif 'post-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                else:
                    last_post_url = url / post_num_str
                await write_last_post_file(self.output_last[1], str(last_post_url))

            return

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url))
            logger.debug(e)



import asyncio

from .scraper_helper import ScrapeMapper
from ..base_functions.base_functions import log
from ..base_functions.data_classes import AuthData, SkipData
from ..client.client import Client


async def scrape(urls, client: Client, include_id: bool, xbunker_auth: AuthData, socialmediagirls_auth: AuthData,
                 simpcity_auth: AuthData, separate_posts: bool, skip_data: SkipData, output_last: list):
    await log("Starting Scrape")

    scraper = ScrapeMapper(client=client, include_id=include_id, xbunker_auth=xbunker_auth,
                           socialmediagirls_auth=socialmediagirls_auth, simpcity_auth=simpcity_auth,
                           separate_posts=separate_posts, skip_data=skip_data, output_last=output_last)
    tasks = []
    for link in urls:
        tasks.append(scraper.map_url(link))
    await asyncio.gather(*tasks)

    Cascade = scraper.Cascade
    await Cascade.dedupe()
    return Cascade



import aiofiles
import logging
import os
import re
from pathlib import Path

from colorama import Fore, Style
from yarl import URL

"""This file contains generic information and functions that are used around the program"""

FILE_FORMATS = {
    'Images': {
        '.jpg', '.jpeg', '.png', '.gif',
        '.webp', '.jpe', '.svg', '.jfif',
        '.tif', '.tiff', '.jif',
    },
    'Videos': {
        '.mpeg', '.avchd', '.webm', '.mpv',
        '.swf', '.avi', '.m4p', '.wmv',
        '.mp2', '.m4v', '.qt', '.mpe',
        '.mp4', '.flv', '.mov', '.mpg',
        '.ogg', '.mkv', '.mts', '.ts'
    },
    'Audio': {
        '.mp3', '.flac', '.wav', '.m4a'
    },
    'Other': {
        '.json', '.torrent', '.zip', '.rar', '.7z', '.torrent'
    }
}

MAX_FILENAME_LENGTH = 100

logger = logging.getLogger(__name__)


async def sanitize(name: str) -> str:
    return re.sub(r'[<>:"/\\|?*\']', "", name).strip()


async def log(text, style=Fore.WHITE) -> None:
    logger.debug(text)
    print(style + str(text) + Style.RESET_ALL)


async def clear() -> None:
    """Clears the terminal screen"""
    os.system('cls' if os.name == 'nt' else 'clear')


async def make_title_safe(title: str):
    title = re.sub(r'[\\*?:"<>|./]', "-", title)
    return title


async def purge_dir(dirname, in_place=True):
    deleted = []
    dir_tree = list(os.walk(dirname, topdown=False))

    for tree_element in dir_tree:
        sub_dir = tree_element[0]
        dir_count = len(os.listdir(sub_dir))
        if dir_count == 0: # Helps with readability and i've had issues with it deleting non-empty dirs
            deleted.append(sub_dir)

    if in_place:
        list(map(os.rmdir, deleted))


async def regex_links(urls) -> list:
    all_links = [x.group().replace(".md.", ".") for x in re.finditer(
        r"(?:http.*?)(?=('|$|\n|\r\n|\r|\s|\"|\[/URL]|]\[|\[/img]))", urls)]
    yarl_links = []
    for link in all_links:
        yarl_links.append(URL(link))
    return yarl_links


async def cyberdrop_parse(url: URL) -> URL:
    mapping_direct = [r'img-...cyberdrop...',
                      r'f.cyberdrop...', r'fs-...cyberdrop...']
    url = str(url)
    for mapping in mapping_direct:
        url = re.sub(mapping, 'cyberdrop.to', url)
    return URL(url)


async def write_last_post_file(file: Path, url: str):
    async with aiofiles.open(file, mode='a') as f:
        await f.write(url + '\n')
    return


async def check_direct(url: URL):
    mapping_direct = ['i.pixl.is', r's..putmega.com', r's..putme.ga', r'img-...cyberdrop...', r'f.cyberdrop...',
                      r'fs-...cyberdrop...', r'cdn.bunkr...', r'cdn..bunkr...', r'media-files.bunkr...', r'jpg.church/images/...',
                      r'simp..jpg.church']
    return any(re.search(domain, url.host) for domain in mapping_direct)



import asyncio
from dataclasses import dataclass
from typing import ClassVar, Dict, List, Optional, Tuple

from yarl import URL


@dataclass
class FileLock:
    locked_files = []

    async def check_lock(self, filename):
        await asyncio.sleep(.1)
        return filename.lower() in self.locked_files

    async def add_lock(self, filename):
        self.locked_files.append(filename.lower())

    async def remove_lock(self, filename):
        self.locked_files.remove(filename.lower())


@dataclass
class AlbumItem:
    """Class for keeping track of download links for each album"""
    title: str
    link_pairs: List[Tuple]
    password: Optional[str] = None

    async def add_link_pair(self, link, referral):
        self.link_pairs.append((link, referral))

    async def set_new_title(self, new_title: str):
        self.title = new_title


@dataclass
class DomainItem:
    domain: str
    albums: Dict[str, AlbumItem]

    async def add_to_album(self, title: str, link: URL, referral: URL):
        if link:
            if title in self.albums.keys():
                await self.albums[title].add_link_pair(link, referral)
            else:
                self.albums[title] = AlbumItem(title=title, link_pairs=[(link, referral)])

    async def add_album(self, title: str, album: AlbumItem):
        if title in self.albums.keys():
            stored_album = self.albums[title]
            for link_pair in album.link_pairs:
                link, referral = link_pair
                await stored_album.add_link_pair(link, referral)
        else:
            self.albums[title] = album

    async def append_title(self, title):
        if not title:
            return
        new_albums = {}
        for album_str, album in self.albums.items():
            new_title = title+'/'+album_str
            new_albums[new_title] = album
            album.title = new_title
        self.albums = new_albums


@dataclass
class CascadeItem:
    """Class for keeping track of domains for each scraper type"""
    domains: Dict[str, DomainItem]

    async def add_albums(self, domain_item: DomainItem):
        domain = domain_item.domain
        albums = domain_item.albums
        for title, album in albums.items():
            await self.add_album(domain, title, album)

    async def add_to_album(self, domain: str, title: str, link: URL, referral: URL):
        if domain in self.domains.keys():
            await self.domains[domain].add_to_album(title, link, referral)
        else:
            self.domains[domain] = DomainItem(domain, {title: AlbumItem(title, [(link, referral)])})

    async def add_album(self, domain: str, title: str, album: AlbumItem):
        if domain in self.domains.keys():
            await self.domains[domain].add_album(title, album)
        else:
            self.domains[domain] = DomainItem(domain, {title: album})

    async def is_empty(self):
        for _, domain in self.domains.items():
            for _, album in domain.albums.items():
                if album.link_pairs:
                    return False
        return True

    async def append_title(self, title):
        if not title:
            return
        for _, domain in self.domains.items():
            new_albums = {}
            for album_str, album in domain.albums.items():
                new_title = title+'/'+album_str
                new_albums[new_title] = album
                album.title = new_title
            domain.albums = new_albums

    async def extend(self, Cascade):
        if Cascade:
            if Cascade.domains:
                for domain_str, domain in Cascade.domains.items():
                    for album_str, album in domain.albums.items():
                        await self.add_album(domain_str, album_str, album)

    async def dedupe(self):
        for _, domain in self.domains.items():
            for _, album in domain.albums.items():
                check = []
                allowed = []
                for pair in album.link_pairs:
                    url, _ = pair
                    if url in check:
                        continue
                    check.append(url)
                    allowed.append(pair)
                album.link_pairs = allowed


@dataclass
class AuthData:
    """Class for keeping username and password"""
    username: str
    password: str


@dataclass
class SkipData:
    supported_hosts: ClassVar[Tuple[str]] = (
        "anonfiles", "bayfiles", "bunkr", "coomer.party", "cyberdrop", "cyberfile", "erome", "gfycat", "gofile", "img.kiwi",
        "jpg.church", "jpg.homes", "kemono.party", "pixeldrain", "pixl.is", "postimg.cc", "putme.ga",
        "putmega.com", "redgifs", "saint", "socialmediagirls", "simpcity", "xbunker")
    sites: List[str]



import asyncio
import json
import logging
import ssl
from pathlib import Path

import aiofiles
from bs4 import BeautifulSoup
from yarl import URL

import aiohttp
import certifi
from tqdm import tqdm

from .rate_limiting import AsyncRateLimiter, throttle
from ..base_functions.base_functions import logger
from ..base_functions.data_classes import FileLock


class Client:
    def __init__(self, ratelimit: int, throttle: int):
        self.ratelimit = ratelimit
        self.throttle = throttle
        self.simultaneous_session_limit = asyncio.Semaphore(50)
        self.user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896.75 Safari/537.36'
        self.ssl_context = ssl.create_default_context(cafile=certifi.where())
        self.cookies = aiohttp.CookieJar(quote_cookie=False)


class Session:
    def __init__(self, client: Client) -> None:
        self.client = client
        self.rate_limiter = AsyncRateLimiter(self.client.ratelimit)
        self.headers = {"user-agent": client.user_agent}
        self.timeouts = aiohttp.ClientTimeout(5 * 60, 15)
        self.client_session = aiohttp.ClientSession(headers=self.headers, raise_for_status=True,
                                                    cookie_jar=self.client.cookies, timeout=self.timeouts)

    async def get_BS4(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    text = await response.text()
                    soup = BeautifulSoup(text, 'html.parser')
                    return soup

    async def get_text(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    text = await response.text()
                    return text

    async def get_json(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    content = json.loads(await response.content.read())
                    return content

    async def post_no_resp(self, url: URL, headers: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context) as response:
                    pass

    async def post_data_no_resp(self, url: URL, data: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.post(url, data=data, headers=self.headers, ssl=self.client.ssl_context) as response:
                    pass

    async def post(self, url: URL, data: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.post(url, data=data, headers=self.headers, ssl=self.client.ssl_context) as response:
                    content = json.loads(await response.content.read())
                    return content

    async def exit_handler(self):
        try:
            await self.client_session.close()
        except Exception as e:
            logging.debug(f"Failed to close sqlite database connection: {str(e)}")


class DownloadSession:
    def __init__(self, client: Client, conn_timeout: int):
        self.client = client
        self.headers = {"user-agent": client.user_agent}
        self.timeouts = aiohttp.ClientTimeout(5*60, conn_timeout)
        self.client_session = aiohttp.ClientSession(headers=self.headers, raise_for_status=True,
                                                    cookie_jar=self.client.cookies, timeout=self.timeouts)
        self.throttle_times = {}

    async def get_filename(self, url: URL, referer: str, current_throttle: int):
        headers = {'Referer': referer, 'user-agent': self.client.user_agent}
        await throttle(self, current_throttle, url.host)
        async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context,
                                           raise_for_status=True) as resp:
            filename = resp.content_disposition.filename
            return filename

    async def get_filesize(self, url: URL, referer: str, current_throttle: int):
        headers = {'Referer': referer, 'user-agent': self.client.user_agent}
        await throttle(self, current_throttle, url.host)
        async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context,
                                           raise_for_status=True) as resp:
            total_size = int(resp.headers.get('Content-Length', str(0)))
            return total_size

    async def get_content_type(self, url: URL, referer: str, current_throttle: int):
        headers = {'Referer': referer, 'user-agent': self.client.user_agent}
        await throttle(self, current_throttle, url.host)
        async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context,
                                           raise_for_status=True) as resp:
            content_type = resp.headers.get('Content-Type')
            return content_type.lower()

    async def download_file(self, url: URL, referer: str, current_throttle: int, range: str, original_filename: str,
                            filename: str, temp_file: str, resume_point: int, show_progress: bool,
                            File_Lock: FileLock, folder: Path, title: str):
        headers = {'Referer': referer, 'user-agent': self.client.user_agent}
        if range:
            headers['Range'] = range
        await throttle(self, current_throttle, url.host)
        async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context,
                                           raise_for_status=True) as resp:
            content_type = resp.headers.get('Content-Type')
            if 'text' in content_type.lower() or 'html' in content_type.lower():
                logger.debug("Server for %s is either down or the file no longer exists", str(url))
                await File_Lock.remove_lock(original_filename)
                return

            total = int(resp.headers.get('Content-Length', str(0))) + resume_point
            (folder / title).mkdir(parents=True, exist_ok=True)

            with tqdm(total=total, unit_scale=True, unit='B', leave=False, initial=resume_point, desc=filename,
                      disable=(not show_progress)) as progress:
                async with aiofiles.open(temp_file, mode='ab') as f:
                    async for chunk, _ in resp.content.iter_chunks():
                        await asyncio.sleep(0)
                        await f.write(chunk)
                        progress.update(len(chunk))



from typing import Dict

from yarl import URL

from ..base_functions.base_functions import log, logger
from ..base_functions.data_classes import DomainItem
from ..client.client import Session


class PostImgCrawler:
    def __init__(self, *, include_id=False):
        self.include_id = include_id

    async def fetch(self, session: Session, url: URL):
        domain_obj = DomainItem(url.host, {})
        await log("Starting scrape of " + str(url))

        try:
            if "gallery" in url.parts:
                listed = await self.folder(session, url)
                for referer, img in listed.items():
                    await domain_obj.add_to_album(url.raw_name, img, referer)
            else:
                img = await self.singular(session, url)
                await domain_obj.add_to_album("Loose PostIMG Files", img, url)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url))
            logger.debug(e)

        await log("Finished scrape of " + str(url))

        return domain_obj

    async def folder(self, session: Session, url: URL):
        album = url.raw_name
        data = {"action": "list", "album": album}
        output = {}
        i = 1
        while True:
            data_used = data
            data_used["page"] = i
            data_out = await session.post(URL("https://postimg.cc/json"), data_used)
            if data_out['status_code'] != 200 or not data_out['images']:
                break
            for item in data_out['images']:
                output[URL("https://postimg.cc/" + item[0])] = URL(item[4].replace(item[0], item[1]))
            i += 1
        return output

    async def singular(self, session: Session, url: URL):
        soup = await session.get_BS4(url)
        link = URL(soup.select_one("a[id=download]").get('href').replace("?dl=1", ""))
        return link



import argparse
import multiprocessing
import platform
from argparse import Namespace
import asyncio
import logging
from asyncio.proactor_events import _ProactorBasePipeTransport
from pathlib import Path
from functools import wraps

from colorama import Fore
from yarl import URL

from . import __version__ as VERSION
from cyberdrop_dl.base_functions.base_functions import clear, log, logger, purge_dir, regex_links
from cyberdrop_dl.base_functions.data_classes import AuthData, SkipData
from cyberdrop_dl.base_functions.sql_helper import SQLHelper
from cyberdrop_dl.client.client import Client
from cyberdrop_dl.client.downloaders import get_downloaders
from cyberdrop_dl.scraper.scraper import scrape


def parse_args():
    parser = argparse.ArgumentParser(description="Bulk downloader for multiple file hosts")
    parser.add_argument("-V", "--version", action="version", version="%(prog)s " + VERSION)
    parser.add_argument("-i", "--input-file", type=Path, help="file containing links to download", default="URLs.txt")
    parser.add_argument("-o", "--output-folder", type=Path, help="folder to download files to", default="Downloads")
    parser.add_argument("--log-file", help="log file to write to", default="downloader.log")
    parser.add_argument("--db-file", help="history database file to write to", default="download_history.sqlite")
    parser.add_argument("--threads", type=int, help="number of threads to use (0 = max)", default=0)
    parser.add_argument("--attempts", type=int, help="number of attempts to download each file", default=10)
    parser.add_argument("--connection-timeout", type=int, help="number of seconds to wait attempting to connect to a URL during the downloading phase", default=15)
    parser.add_argument("--disable-attempt-limit", help="disables the attempt limitation", action="store_true")
    parser.add_argument("--include-id", help="include the ID in the download folder name", action="store_true")
    parser.add_argument("--exclude-videos", help="skip downloading of video files", action="store_true")
    parser.add_argument("--exclude-images", help="skip downloading of image files", action="store_true")
    parser.add_argument("--exclude-audio", help="skip downloading of audio files", action="store_true")
    parser.add_argument("--exclude-other", help="skip downloading of images", action="store_true")
    parser.add_argument("--ignore-history", help="This ignores previous download history", action="store_true")
    parser.add_argument("--output-last-forum-post", help="Separates forum scraping into folders by post number", action="store_true")
    parser.add_argument("--separate-posts", help="Separates forum scraping into folders by post number", action="store_true")
    parser.add_argument("--xbunker-username", type=str, help="username to login to xbunker", default=None)
    parser.add_argument("--xbunker-password", type=str, help="password to login to xbunker", default=None)
    parser.add_argument("--socialmediagirls-username", type=str, help="username to login to socialmediagirls", default=None)
    parser.add_argument("--socialmediagirls-password", type=str, help="password to login to socialmediagirls", default=None)
    parser.add_argument("--simpcity-username", type=str, help="username to login to simpcity", default=None)
    parser.add_argument("--simpcity-password", type=str, help="password to login to simpcity", default=None)
    parser.add_argument("--skip", dest="skip_hosts", choices=SkipData.supported_hosts, help="This removes host links from downloads", action="append", default=[])
    parser.add_argument("--ratelimit", type=int, help="this will add a ratelimiter to requests made in the program during scraping, the number you provide is in requests/seconds", default=50)
    parser.add_argument("--throttle", type=int, help="This is a throttle between requests during the downloading phase, the number is in seconds", default=0.5)
    parser.add_argument("links", metavar="link", nargs="*", help="link to content to download (passing multiple links is supported)", default=[])
    args = parser.parse_args()
    return args


async def download_all(args: argparse.Namespace):
    await clear()
    await log(f"We are running version {VERSION} of Cyberdrop Downloader", Fore.WHITE)
    print_args = Namespace(**vars(args)).__dict__
    print_args['xbunker_password'] = '!REDACTED!'
    print_args['socialmediagirls_password'] = '!REDACTED!'
    print_args['simpcity_password'] = '!REDACTED!'
    logging.debug(f"Starting downloader with args: {print_args}")
    input_file = args.input_file
    if not input_file.is_file():
        input_file.touch()
        await log(f"{input_file} created. Populate it and retry.")
        exit(1)

    client = Client(args.ratelimit, args.throttle)
    SQL_helper = SQLHelper(args.ignore_history, args.db_file)
    await SQL_helper.sql_initialize()

    threads = args.threads if args.threads != 0 else multiprocessing.cpu_count()

    links = args.links
    links = list(map(URL, links))

    with open(input_file, "r", encoding="utf8") as f:
        links += await regex_links(f.read())

    links = list(filter(None, links))

    if not links:
        await log("No links found, check the URL.txt\nIf the link works in your web browser, "
                  "please open an issue ticket with me.", Fore.RED)

    output_url_file = None

    if args.output_last_forum_post:
        output_url_file: Path = input_file.parent / "URLs_last_post.txt"
        if output_url_file.exists():
            output_url_file.unlink()
            output_url_file.touch()

    xbunker_auth = AuthData(args.xbunker_username, args.xbunker_password)
    socialmediagirls_auth = AuthData(args.socialmediagirls_username, args.socialmediagirls_password)
    simpcity_auth = AuthData(args.simpcity_username, args.simpcity_password)
    skip_data = SkipData(args.skip_hosts)
    excludes = {'videos': args.exclude_videos, 'images': args.exclude_images, 'audio': args.exclude_audio,
                'other': args.exclude_other}
    content_object = await scrape(links, client, args.include_id, xbunker_auth, socialmediagirls_auth,
                                  simpcity_auth, args.separate_posts, skip_data, [args.output_last_forum_post, output_url_file])

    if await content_object.is_empty():
        logging.error('ValueError No links')
        await log("No links found duing scraping, check passwords or that the urls are accessible", Fore.RED)
        await log("This program does not currently support password protected albums.", Fore.RED)
        exit(0)
    await clear()

    downloaders = await get_downloaders(content_object, folder=args.output_folder, attempts=args.attempts,
                                        disable_attempt_limit=args.disable_attempt_limit, max_workers=threads,
                                        excludes=excludes, SQL_helper=SQL_helper, client=client)

    for downloader in downloaders:
        await downloader.download_content(conn_timeout=args.connection_timeout)
    logger.debug("Finished")

    partial_downloads = [str(f) for f in args.output_folder.rglob("*.part") if f.is_file()]

    await log('Purging empty directories')
    await purge_dir(args.output_folder)

    await log('Finished downloading. Enjoy :)')
    if partial_downloads:
        await log('There are still partial downloads in your folders, please re-run the program.')


def silence_event_loop_closed(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except RuntimeError as e:
            if str(e) != 'Event loop is closed':
                raise
    return wrapper


def main(args=None):
    if not args:
        args = parse_args()
    logging.basicConfig(
        filename=args.log_file,
        level=logging.DEBUG,
        format="%(asctime)s:%(levelname)s:%(module)s:%(filename)s:%(lineno)d:%(message)s",
        filemode="w"
    )

    if platform.system() == 'Windows':
        # Silence the "Event loop is closed" exception here.
        _ProactorBasePipeTransport.__del__ = silence_event_loop_closed(_ProactorBasePipeTransport.__del__)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(download_all(args))
    loop.run_until_complete(asyncio.sleep(1))
    loop.close()


if __name__ == '__main__':
    print("""
    STOP! If you're just trying to download files, check the README.md file for instructions.
    If you're developing this project, use start.py instead.
    """)
    exit()
